// TimerPodule1.cpp: implementation of the CTimerPodule class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "resource.h"

// unfortunate clash...
#undef LoadString

#include "Console.h"
#include "UserInterface.h"
#include "TimerPodule1.h"
#include "TimerProperties.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>



// callback rate in hz
#define TIMER_RATE	100

// registers

// read/write timer register. Incremented every 10ms by a
// callback timer
#define TREG_TIMER				0x2000

// read returns the interrupt status
// write sets the interrupt mask and clears any interrupt
#define TREG_CLEARINTERRUPT		0x2004


// page register for the podule rom
#define TREG_ROMPAGE			0x2008

// rom is accessed on word boundaries and spread into 0x2000 bytes
// ie half of the module workspace
#define ROM_PAGE_SHIFT 11
#define ROM_PAGE_SIZE (1<<ROM_PAGE_SHIFT)
#define ROM_PAGE_MASK (ROM_PAGE_SIZE-1)

// single static instance of timer properties
static CTimerProperties* ppage = 0;

// declare the component so it can be created from a model file
IMPLEMENT_COMPONENT("TimerPodule", CTimerPodule);

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

// callback handler for timer event.
static Word timer_event(IEvent*e, int delay)
{
	return ((CTimerPodule*)e->iface.impl)->TimerEvent(delay);
}


// Constructor. Only basic initialisation should be done here
CTimerPodule::CTimerPodule()
{
	// Only simple initialisation in here
	rombase = 0;
	romsize = 0;
	rompage = 0;

	timer.iface.size = sizeof(IEvent);
	timer.iface.type = "ExampleTimer";	// not exported, just give it a friendly name
	timer.iface.impl = this;
	timer.ProcessEvent = timer_event;

	int_mask = 0;
	int_status = 0;
}

CTimerPodule::~CTimerPodule()
{

}

// In init we perform local tasks such as allocating memory
BOOL CTimerPodule::Init()
{
	// Load the rom
	const char* romname = LoadString("rom", "timerrom.dat");
	char path[1024];

	FILE* fp = 0;
	for (const char** spath = CComponentManager::GetSearchPath(); *spath; ++spath)
	{
		strcpy(path, *spath);
		strcat(path, romname);

		fp = fopen(path, "rb");
		if (fp)
			break;
	}
	if (fp)
	{
		fseek(fp, 0, SEEK_END);
		romsize = ftell(fp);
		fseek(fp, 0, SEEK_SET);

		rombase = (Byte*)malloc(romsize);
		fread(rombase, 1, romsize, fp);
		fclose(fp);
	}
	else
	{
		console->printf("TimerPodule failed to  load rom image\n");
		return FALSE;
	}

	// Create a single instance of the property page
	if (!ppage)
	{
		ppage = new CTimerProperties;
		ppage->podule = this;
	}
	return CIocPodule::Init();
}

// In connect we connect to other components
BOOL CTimerPodule::Connect()
{
	// add the timer

	scheduler = (IScheduler*)FindInterface(IFACE_SCHEDULER);
	if (!scheduler)
		return FALSE;

	timer_ticks = scheduler->TickRate(scheduler) / TIMER_RATE;
	scheduler->ScheduleEvent(scheduler, &timer, timer_ticks);

	// Export the property page
	IWinUI *ui = (IWinUI*)FindInterface(IFACE_USERINTERFACE);
	if (ui)
		ui->AddPropertyPage(ui, ppage);

	return CIocPodule::Connect();
}

BOOL CTimerPodule::Disconnect()
{
	IWinUI *ui = (IWinUI*)FindInterface(IFACE_USERINTERFACE);
	if (ui)
		ui->RemovePropertyPage(ui, ppage);

	if (scheduler)
		scheduler->DescheduleEvent(scheduler, &timer);
	
	scheduler = 0;

	return CIocPodule::Disconnect();
}

BOOL CTimerPodule::Cleanup()
{
	free(rombase);
	rombase = 0;
	romsize = 0;
	rompage = 0;

	
	return CIocPodule::Cleanup();
}

// Handle byte writes to our podule.
void CTimerPodule::WriteByte(Word address, Byte b)
{
	if (address == TREG_TIMER)
	{
		ticks = b;
	}
	else if (address == TREG_CLEARINTERRUPT)
	{
		int_status = 0;
		int_mask = b;
		ClearIrq();
	}
	else if (address == TREG_ROMPAGE)
	{
		if ((b<<ROM_PAGE_SHIFT) < romsize)
			rompage = b;
		else
			rompage = 0;
	}
}

// Handle word writes to our podule. Only the timer
// is 32bit
void CTimerPodule::WriteWord(Word address, Word w)
{
	if (address == TREG_TIMER)
		ticks = w;
	else
		WriteByte(address, (Byte)w);
}

// Handle byte reads from our podule
Byte CTimerPodule::ReadByte(Word address)
{
	if (address == TREG_TIMER)
	{
		return ticks;
	}
	else if (address == TREG_CLEARINTERRUPT)
	{
		return int_status;
	}
	else if (rombase)
	{
		Word romaddr = ((address>>2)&ROM_PAGE_MASK) + (rompage<<ROM_PAGE_SHIFT);
		return rombase[romaddr];
	}
	else
		return 0xff;
}

// handle word reads from our podule
Word CTimerPodule::ReadWord(Word address)
{
	if (address == TREG_TIMER)
	{
		return ticks;
	}
	else
		return ReadByte(address);
}

// A timer event happend. Just signal an interrupt and carry on
int CTimerPodule::TimerEvent(int overrun)
{
	ticks++;

	int_status = int_mask;

	// Don't enable interrupts unless we're ready to handle them
	if (int_status)
		SetIrq();

	// carry on at the same rate
	return timer_ticks;
}
