// PluginBusDevice.h: interface for the CPluginBusDevice class.
//
// Proxy class which passes calls on via a PlgComponent interface
// to the equivalent CPluginBusDevice of the given class.
//
//////////////////////////////////////////////////////////////////////
#if !defined(BUSINTERFACEPUBLIC_H)
#define BUSINTERFACEPUBLIC_H

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Interface.h"

#define IFACE_PHSICALBUS "PhysicalBus"

// macros to read words from physical pages
#define WRITEW(base, offset, w) (*((Word*)((Byte*)(base) + (offset))) = (w))
#define READW(base, offset) (*((Word*)((Byte*)(base) + (offset))))

#define WRITEB(base, offset, w) (*((Byte*)(base) + (offset)) = (w))
#define READB(base, offset) (*((Byte*)(base) + (offset)))

#define WRITEH(base, offset, w) (*((HalfWord*)((Byte*)(base) + (offset))) = (w))
#define READH(base, offset) (*((HalfWord*)( ((Byte*)(base) + (offset)) )))

// standard physical bus constants
#define PPAGE_SHIFT 12
#define PPAGE_SIZE  (1<<PPAGE_SHIFT)
#define PPAGE_MASK	(PPAGE_SIZE-1)

#define INVALID_PPAGE 0xffffffff

// interface implemented by physical bus devices. It doesn't
// need to be exported. This will be modified by the PhysicalBus class
struct IPhysicalBusDevice
{
	Interface iface;

	// If this device is acting as a filter or doesn't respond to
	// reads or writes, this is the next lower device in the chain.
	// E.g. ROM doesn't respond to writes so it will return the
	// result of writing to the next device. The PhysicalBus will catch
	// reads/writes at the bottom of a device stack
	IPhysicalBusDevice* next_device; 

	// called when the arm writes to a memory location mapped to this
	// device. The functions may be NULL.
	void	(*WriteWord)(IPhysicalBusDevice*, Word page, Word offset, Word data);
	void	(*WriteHalf)(IPhysicalBusDevice*, Word page, Word offset, HalfWord data);
	void	(*WriteByte)(IPhysicalBusDevice*, Word page, Word offset, Byte data);

	// called when the arm reads from a memory location mapped to this
	// device. The functions may be NULL.
	Word	 (*ReadWord)(IPhysicalBusDevice*, Word page, Word offset);
	HalfWord (*ReadHalf)(IPhysicalBusDevice*, Word page, Word offset);
	Byte	 (*ReadByte)(IPhysicalBusDevice*, Word page, Word offset);

	// If a page contains ram or rom rather than registers and
	// is effectively cacheable then returning a pointer to the
	// page via these functions will allow the arm cache to read
	// directly from the ram rather than calling the above functions.
	// If the functions are NULL or a NULL pointer is returned the
	// page will not be cached and the above read/write functions will
	// always be called
	Byte* (*ProbeReadCacheable)(IPhysicalBusDevice*, Word page);
	Byte* (*ProbeWriteCacheable)(IPhysicalBusDevice*, Word page);

};

// Interface exported by the physical bus implementation for use by
// bus device implementations
struct IPhysicalBus
{
	Interface iface;

	// Map a bus device to a region of the bus space.
	// If there is already a device at that location then it will be hidden
	// unless the new device passes calls down (i.e. it is a filter). If there
	// is more than one underlying device this call will fail. Thus filters
	// in general should only be applied to one page.
	BOOL (*MapDevice)(IPhysicalBus*, IPhysicalBusDevice*, Word page, Word npages);
	BOOL (*UnmapDevice)(IPhysicalBus*, IPhysicalBusDevice*, Word page, Word npages);

	// Inform bus masters that a previously cacheable page is no longer
	// cacheable or that the contents may have changed.
	void (*PurgePageReadable)(IPhysicalBus*, Word page);
	void (*PurgePageWriteable)(IPhysicalBus*, Word page);
};



/*
e.g.

void rom_write_word(IPhysicalBusDevice* dev, Word page, Word offset, Word data)
{
	(dev->next_device)->WriteWord(dev->next_device, page, offset, data);
}

The class version will go

SimpleDevice::WriteWord(Word page, Word offset, Word data)
{
	(iface.next_device)->WriteWord(dev->next_device, page, offset, data);
}

*/

#endif // BUSINTERFACEPUBLIC
