// Component.h: 
//
// C++ wrappers for the underlying Interface mechanism.
//
// These classes are used throughout the emulator to create
// implementations of various bus interfaces and also for most
// components. You don't _have_ to use them but it is much easier
// to do so.
// 
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_COMPONENT_H__60CA23E0_5FFF_43EA_9E60_5452012486C9__INCLUDED_)
#define AFX_COMPONENT_H__60CA23E0_5FFF_43EA_9E60_5452012486C9__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Interface.h"

class CComponentClass;
class CComponent  
{
public:
	IComponent iface;		// the component interface

protected:
	CComponent();

public:
	virtual CComponentClass* GetClass() { return NULL; }

public:
	virtual ~CComponent();

	// implementation of the IComponent interface. Derived versions may
	// override this Ioctl but the base class does most of the decoding
	// and calls the methods below so it usually isn't necessary
	virtual Word Ioctl(Word op, void* in, void* out);


	// Decoded ioctl interface. Override these methods to perform
	// various actions during the construction and running of
	// a model.

	// Initialise the members of this component, allocate memory
	// create other slave components and register them. Once slave
	// components have been registered they will receive the Init()
	// themselves.
	// Load status from the config file
	virtual BOOL Init()			{ return TRUE; }

	// Fetch other components interfaces and make intercomponent
	// connections.
	virtual BOOL Connect()		{ return TRUE; }

	// remove connections from other components.
	virtual BOOL Disconnect()	{ return TRUE; }
	
	// free up memory, destroy slave components. Slave components
	// will have received the Cleanup() call
	virtual BOOL Cleanup()		{ return TRUE; }

	// start any slave threads if necessary
	virtual BOOL Start()		{ return TRUE; }

	// stop any slave threads
	virtual BOOL Stop()			{ return TRUE; }

	// Reset internal state to a known startup condition
	virtual BOOL Reset()		{ return TRUE; }

	// Set any necessary state in other components
	virtual BOOL PostReset()	{ return TRUE; }

	// Check if a component is OK to quit. The user can overide but
	// they will be asked
	virtual BOOL OkToQuit()		{ return TRUE; }

	// Helper methods:

	// Exports an interface implemented by this component.
	// The argument must be a structure with a first element being
	// an Interface. Only bus style interfaces need be exported.
	// returns the instance number of the exported interface (generally 0)
	// or -1 if there is a problem
	static Word ExportInterface(void* iface);

	// find an interface exported by another component. E.g.
	// Ioc or the PhysicalBus. If an interface is found it can
	// be cast to the correct type to access its methods.
	static Interface* FindInterface(const char *name, int instance = 0);

	// Add a component to the model at runtime. This enables components
	// to be built from other more generic components. This should only
	// be done inside Component::Init()
	static BOOL AddComponent(CComponent*);

	// Remove a runtime created component. This should only be done
	// inside Component::Cleanup()
	static BOOL RemoveComponent(CComponent*);

	// Access settings in persistent storage related to this component
	// It is best to only call these during Init(). Note that only
	// Named components may uses these methods
	void StoreString(const char* key, const char* val);
	const char* LoadString(const char* key, const char* deflt = "");
	void StoreInt(const char* key, Word val);
	Word LoadInt(const char* key, Word deflt = 0);
};

class CComponentClass;

typedef CComponent* (*ComponentCreateFun)();

// This is an implementation of the IComponentClass interface.
// It is used by the DECLARE_COMPONENT and IMPLEMENT_COMPONENT macros
// to register an IComponentClass for each CComponent so the component
// can be dynamically created
class CComponentClass
{
public:

	IComponentClass iface;			// the class interface
	ComponentCreateFun creator;		// function to return a new component of this class

	static CComponentClass *class_list;	// used to build a list of local classes
	CComponentClass *next_class;	// next class in local list

public:
	CComponentClass(ComponentCreateFun creator, const char *name);

	IComponent* CreateComponent();
	static void RegisterClasses();
};

// Macro's to simplify dynamic component definitions
#define DECLARE_COMPONENT(classname)				\
public:												\
	static CComponentClass class##classname;		\
	static CComponent* CreateComponent();			\
	virtual CComponentClass* GetClass()				\
			{ return &class##classname; }			\

#define IMPLEMENT_COMPONENT(comname, classname)			\
	CComponent* classname::CreateComponent()			\
			{ CComponent* c = new classname;			\
			  c->iface.isa = &c->GetClass()->iface; return c; }		\
    CComponentClass classname::class##classname			\
			(classname::CreateComponent, comname);		\


#define COMPONENT_CLASS(classname) (&classname::class##classname)

////////////////////////////////////////////////////////////////////////////////////
//
// Wrapper class for the global component interface. There will be one of these
// per module if this library is used.
//
// INCORPORATE INTO CCOMPONENT AS STATIC METHODS
class CComponentManager
{
private:
	CComponentManager();

public:

	// Register a component class so it can be used to instance a component
	static Word RegisterClass(CComponentClass*);

	// Register a public interface so other components can connect to it
	// If global is true, the interface will exist in ALL models and persist
	// throughout the invocation of the emulator. 
	static Word RegisterInterface(Interface*, BOOL global=FALSE);

	// Register a global component which will exist in all models
	static Word AddGlobalComponent(CComponent*);
	static Word RemoveGlobalComponent(CComponent*);

	// Register a component so it can be built correctly
	// This only needs to be done for run time created components, not
	// those specified in a model file. Persistent components will
	// be added to the model file.
	static Word AddComponent(CComponent*, BOOL persistent=FALSE);
	static Word RemoveComponent(CComponent*);

	// Find a public interface exported by a component. Note that this
	// returns an IComponent interface, not a CComponent.
	static Interface* FindInterface(const char* name, Word instance=0);

	// Find a component. This is currently only useful for setting
	// component properties. 
	static IComponent* FindComponent(const char* name, int instance);

	// Load/store state related to a particular component (not interface)
	// Maybe we'll put these in a a generic Ioctl as there will be more at
	// a future date
	static void StoreString(CComponent* comp, const char* key, const char* val);
	static const char* LoadString(CComponent* comp, const char* key, const char* deflt = "");
	static void StoreInt(CComponent* comp, const char* key, Word val);
	static Word LoadInt(CComponent* comp, const char* key, Word deflt = 0);

	// Need a way to store information in the main config file.


	// Get the search path for auxiliary files. Plugins should search this
	// in order. It is NULL terminated
	static const char** GetSearchPath();

	// Returns some generic information about the application that 
	// loaded this plugin
	static const char* GetAppName();

	// low 8 bits are minor version
	// next 8 bits are major version
	// next 8 bits are application (0=RS, 1=VA)
	// top 8 bits are flags
	//		bit 31=1 - this is a demo version of an app
	static Word GetAppVersion();
};


#endif // !defined(AFX_COMPONENT_H__60CA23E0_5FFF_43EA_9E60_5452012486C9__INCLUDED_)
